# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#

from __future__ import print_function
from __future__ import absolute_import
import json
import os
import sys
import time
import traceback

from typing import Union, Any, Optional, Dict  # NOQA

import psutil

from clcommon import FormattedException
from clcommon.utils import (
    run_command,
    ExternalProgramFailed,
    get_cl_version,
)
from clwizard.config import NoSuchModule
from .config import acquire_config_access
from .config import Config  # NOQA
from .modules import run_installation, ALL_MODULES, get_supported_modules
from .constants import (
    WizardStatus,
    ModuleStatus,
    CRASH_LOG_PATH,
    FILE_MARKER_PATH,
    MAIN_LOG_PATH,
)
from .exceptions import CancelModuleException, InstallationFailedException
from .parser import parse_cloudlinux_wizard_opts
from .utils import (
    is_background_process_running,
    run_background,
    setup_logger,
)
from clcommon.utils import get_rpm_db_errors


class CloudlinuxWizard(object):
    """Main class for working with Wizard that exposes high level logic."""
    # states in which we can remove module from queue
    CANCELABLE_MODULE_STATUSES = [
        ModuleStatus.PENDING,
        ModuleStatus.FAILED,
        ModuleStatus.CANCELLED,
    ]

    # modules states in which wizard modules can be considered as done
    DONE_MODULES_STATUSES = [
        ModuleStatus.INSTALLED,
        ModuleStatus.CANCELLED,
        ModuleStatus.AUTO_SKIPPED
    ]

    def __init__(self):
        self._opts = None
        self._supported_modules = get_supported_modules()
        self.log = setup_logger('wizard.main', MAIN_LOG_PATH)

    def run(self, argv):
        """

        :param argv: command line arguments for wizard
        :return: None
        """
        self._opts = parse_cloudlinux_wizard_opts(argv)
        try:
            if self._opts.subparser == 'install':
                self._validate_system()
                if self.is_installation_finished() and not self._opts.force:
                    self._print_result_and_exit(result='Installation already finished', exit_code=1)
                if self._opts.no_async:
                    run_installation()
                else:
                    self.run_background_installation(options=self._opts.json_data)
            elif self._opts.subparser == 'status':
                self._validate_system()
                if self._opts.initial:
                    self._get_initial_status()
                else:
                    self._get_modules_statuses()
            elif self._opts.subparser == 'cancel':
                self._cancel_module_installation(self._opts.module)
            elif self._opts.subparser == 'finish':
                self.create_completion_marker()
            else:
                raise NotImplementedError

            if (self._opts.subparser in ['install', 'cancel'] and self.is_all_modules_installed()) or \
                    (self._opts.subparser == 'finish' and not self.is_all_modules_installed()):
                '''
                Called only once if:
                    -- in case of install: -all modules was installed successfully
                                           -some module failed during installation, but was installed after resume
                    -- in case of cancel:  -some module failed during installation, but it was canceled by user
                                           and as result all modules in status installed
                    -- in case of finish:  -only if user close wizard when some module in status failed
                '''
                self.run_collecting_statistics()
                self.run_cagefs_force_update()

            self._print_result_and_exit()
        except FormattedException as e:
            self.log.error("We got an error while running cloudlinux-wizard, message: '%s'", str(e))
            self._print_result_and_exit(
                result=e.message, context=e.context, details=e.details, exit_code=1)
        except InstallationFailedException:
            self._print_result_and_exit(
                result="Module installation failed, see log for more information", exit_code=1)
        except Exception as e:
            self.log.exception("Unknown error in cloudlinux-wizard, %s", str(e))
            self._print_result_and_exit(
                result="Unknown error happened, please, try again "
                       "or contact CloudLinux support if it happens again.",
                details=traceback.format_exc())

    @staticmethod
    def is_installation_finished():
        # type: () -> bool
        return os.path.isfile(FILE_MARKER_PATH)

    def create_completion_marker(self):
        # type: () -> None
        try:
            os.mknod(FILE_MARKER_PATH)
            self.log.info("Wizard was successfully finished")
        except (OSError, IOError) as err:
            self.log.warning("Finish command called more than once, error: '%s'", str(err))
            self._print_result_and_exit(result='Finish command called more than once', exit_code=1)

    def run_background_installation(self, options=None):
        # type: (Optional[Dict]) -> Optional[None]
        cmd = sys.argv[:]
        cmd.append('--no-async')
        with acquire_config_access() as config:
            # two processes cannot use config at same time
            # so we can safely do check for running process here
            if is_background_process_running():
                self._print_result_and_exit(
                    result='Unable to run new installation because '
                           'background task is still working', exit_code=1)
            # the only case when options are None is the 'resume' case
            if options is not None:
                config.set_modules(options)
            # worker will not be able to acquire reading lock
            # and will wait unless we finally close config file
            worker_pid = run_background(cmd).pid
            config.worker_pid = worker_pid
        self._print_result_and_exit(result='success', pid=worker_pid)

    def _validate_system(self):
        """
        Check that wizard supports current system
        """
        if get_cl_version() is None:
            self._print_result_and_exit(
                result='Could not identify CloudLinux version. '
                       'Restart your system. If you have the same problem again - '
                       'contact CloudLinux support.'
            )

    def _get_module_log_path(self, module_name):
        """Get path to module log file"""
        return self._supported_modules[module_name].LOG_FILE

    def _get_modules_statuses(self):
        """
        Get information about background worker state
        """
        # we should return modules in order, but config
        # does not know about it, let's sort modules here
        modules = []
        with acquire_config_access() as config:
            state = self._get_wizard_state(config)
            for name in self._supported_modules:
                try:
                    status = config.get_module_status(name)
                    status_time = config.get_module_status_time(name)
                except NoSuchModule:
                    continue
                module_status = {'status': status, 'name': name, 'status_time': status_time}
                if status in [ModuleStatus.FAILED, ModuleStatus.AUTO_SKIPPED]:
                    module_status['log_file'] = self._get_module_log_path(name)
                modules.append(module_status)
        if state == WizardStatus.CRASHED:
            self._print_result_and_exit(wizard_status=state, modules=modules, crash_log=CRASH_LOG_PATH)
        self._print_result_and_exit(wizard_status=state, modules=modules)

    def _get_initial_status(self):
        """
        Get initial modules status that is used
        by lvemanager to display wizard pages
        """
        error_message = get_rpm_db_errors()
        if error_message:
            # RPM DB corrupted
            self._print_result_and_exit(result=error_message)
        self._print_result_and_exit(modules={
            module_name: cls().initial_status()
            for module_name, cls in self._supported_modules.items()
        }, unsuppored_by_cp=list(set(ALL_MODULES) - set(self._supported_modules)))

    def _cancel_module_installation(self, module):
        # type: (str) -> Optional[None]
        """Remove module from queue or print error if it is not possible"""
        self.log.info("Trying to cancel module '%s'", module)
        with acquire_config_access() as config:
            status = config.get_module_status(module)
            if status in self.CANCELABLE_MODULE_STATUSES:
                config.set_module_status(
                    module_name=module,
                    new_state=ModuleStatus.CANCELLED
                )
                self.log.info("Module '%s' successfully canceled", module)
            else:
                self.log.warning("Not able to cancel module '%s', "
                                 "because it is in status '%s'", module, status)
                raise CancelModuleException(module, status)

    def run_collecting_statistics(self):
        """
        Collects user`s statistics
        """
        cmd = ['/usr/sbin/cloudlinux-summary', '--send']
        if not os.environ.get('SYNCHRONOUS_SUMMARY'):
            cmd.append('--async')

        self.log.info('Try to collect statistics..')
        try:
            out = run_command(cmd)
            self.log.info("Output of statistics collection command: '%s'", out)
        except ExternalProgramFailed as e:
            self.log.error("Error during collecting and sending statistics. Reason is '%s'", e)

    def is_all_modules_installed(self):
        # type: () -> bool
        """
        Check that all modules were successfully installed:
        -- installed
        -- canceled
        -- auto-skipped
        """
        with acquire_config_access() as config:
            statuses = list(config.statuses.values())
        return all(status in self.DONE_MODULES_STATUSES for status in statuses)

    def run_cagefs_force_update(self):
        """
        Runs cagefsctl --force-update in background
        """
        cagefsctl_bin = '/usr/sbin/cagefsctl'

        if not os.path.isfile(cagefsctl_bin):
            return
        cmd = [cagefsctl_bin, '--force-update', '--wait-lock']
        self.log.info('Starting cagefs force-update in background..: %s', cmd)
        run_background(cmd)

    def _get_wizard_state(self, config):
        # type: (Config) -> str
        # worker pid is None only in case when wizard
        # wasn't EVER called, and this worker pid will stay
        # in config forever, even after wizard is Done
        if config.worker_pid is None:
            return WizardStatus.IDLE

        try:
            psutil.Process(config.worker_pid)
        except psutil.NoSuchProcess:
            # Background process is no longer alive.
            # 1. Wizard DONE: all modules are in state "installed", "cancelled" or "auto-skipped".
            # 2. Wizard FAILED: one of the modules in state "failed" or "canceled" and no modules in status "installing"
            # 3. Wizard CRASHED: none of the above.
            statuses = list(config.statuses.values())
            if all(status in self.DONE_MODULES_STATUSES for status in statuses):
                return WizardStatus.DONE
            # cancel module`s status is acceptable for general wizard status FAILED, DO NOT CHANGE IT PLS (LU-1295)
            # extra check for installing status is needed to exclude possible wizard status CRASHED
            elif any(status in (ModuleStatus.FAILED, ModuleStatus.CANCELLED) for status in statuses)\
                    and not any(status in (ModuleStatus.INSTALLING, ) for status in statuses):
                return WizardStatus.FAILED
            return WizardStatus.CRASHED
        else:
            return WizardStatus.IN_PROGRESS

    @staticmethod
    def _print_result_and_exit(result='success', exit_code=0, **extra):
        # type: (str, int, **Any) -> Union[None]
        """
        Print data in default format for web and exit
        :param dict extra: extra fields in response,
                           usually we expect 'context' here
        """
        message = {
            'result': result,
            'timestamp': time.time()
        }
        message.update(extra)
        print(json.dumps(message, indent=2, sort_keys=True))
        sys.exit(exit_code)
