# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#

# pylint: disable=no-absolute-import

import os
import signal
import time
import json
import socket
import select
import pwd
import struct
from typing import Optional, List, Dict, Tuple
from contextlib import contextmanager
from collections import namedtuple

from clcommon.cpapi import cpusers
from clcommon.utils import get_process_pid, remove_pid_file, write_pid_file
from clwpos.daemon_redis_lib import is_user_redis_alive, kill_all_users_redises, reload_redis_for_user
import clwpos.socket_utils as socket_utils
from clwpos.daemon_base import WposDaemonBase
from clwpos.constants import WPOS_DAEMON_SOCKET_FILE
from clcommon.utils import is_uid_present


Socket_request = namedtuple('Socket_request', ['client_socket', 'client_uid', 'client_cmd_dict'])


@contextmanager
def _umask_0():
    """
    Context manager for dropping umask
    """
    prev = os.umask(0)
    yield
    os.umask(prev)


def _create_socket() -> socket.socket:
    """
    Create world-writable socket in given sock_location
    or reuse existing one
    :return: socket object
    """
    with _umask_0():
        sockobj = socket.socket(socket.AF_UNIX)
        sockobj.bind(WPOS_DAEMON_SOCKET_FILE)
        sockobj.listen()
        return sockobj


class WposDaemon(WposDaemonBase):
    """
    WPOS daemon main class. Socket operations, redis process control, etc
    """
    _WPOS_DAEMON_RELOAD_FILE = "/var/lve/wpos_reload"
    _DAEMON_RELOAD_COMMAND = 'reload'
    _DAEMON_GET_REDIS_STATUS_COMMAND = 'get-redis-status'
    _DAEMON_VALID_COMMANDS = [_DAEMON_RELOAD_COMMAND, _DAEMON_GET_REDIS_STATUS_COMMAND]

    def __init__(self):
        super().__init__()
        # Dictionary with user's redices PIDs for monitoring
        # Example: {'cltest1': 12345}
        self._monitoring_users_dict = {}
        self._username_list_to_process = []
        self._socket: Optional[socket.socket] = None
        # User uid from socket to reload, 0 - no user to reload
        self._reload_uid = 0
        self._update_monitoring_interval()
        self._logger.info("Cloudlinux WPOS daemon uses monitoring interval: %d seconds", self._monitoring_interval)
        # List of socket queries. Each list item is tuple Socket_request
        self._socket_request_list: List[Socket_request] = []

    def _reload_redis_for_user(self, username: str) -> Dict:
        """
        Starts/stops redis for user and updates dict for monitoring (self._monitoring_users_dict)
        :param username: username to reload redis
        :return: dict {"result": "success"} / {"result": "error", "context": {}}
        """
        # self._monitoring_users_dict example:
        # {'cltest1': 12345}
        try:
            old_redis_pid = self._monitoring_users_dict.get(username)
            new_redis_pid, result_dict = reload_redis_for_user(self._logger, username, old_redis_pid)
            # check error
            if result_dict["result"] != "success":
                return result_dict
            if new_redis_pid != -1:
                # Redis was started for user
                self._monitoring_users_dict[username] = new_redis_pid
            else:
                # Redis was stopped for user
                if username in self._monitoring_users_dict:
                    del self._monitoring_users_dict[username]
        except Exception as e:
            self._logger.exception(e)
            return {"result": str(e)}
        return {"result": "success"}

    def _redises_monitoring(self):
        """
        Working redises monitoring
        """
        userlist_to_remove = []
        for username in self._monitoring_users_dict.keys():
            try:
                is_redis_alive, _ = is_user_redis_alive(self._logger, username)
                if not is_redis_alive:
                    # Redis process is died, restart it
                    redis_pid, _ = reload_redis_for_user(self._logger, username, self._monitoring_users_dict[username])
                    if redis_pid != -1:
                        self._monitoring_users_dict[username] = redis_pid
                    else:
                        userlist_to_remove.append(username)
            except Exception as e:
                self._logger.exception(e)
        # Remove absent users from monitoring list
        for username in userlist_to_remove:
            del self._monitoring_users_dict[username]

    def _process_all_requests(self):
        """
        Process all requests - start/stop redis for users
        """
        if self._reload_config_need:
            # if SIGHUP was happened - reload all users config (full reload)
            self._reload_config_need = False
            # Read all user .clwpos/clwpos_config.json data
            username_list_to_process = list(cpusers())
        else:
            username_list_to_process = []
        # Process socket_requests
        for socket_request_data in self._socket_request_list:
            client_socket, client_uid, client_cmd_dict = socket_request_data
            if client_cmd_dict['command'] == self._DAEMON_RELOAD_COMMAND:
                try:
                    # Reload redis for user
                    username = pwd.getpwuid(client_uid).pw_name
                    if username in username_list_to_process:
                        username_list_to_process.remove(username)
                    result_dict = self._reload_redis_for_user(username)
                    result_dict['timestamp'] = time.time()
                    socket_utils.send_dict_to_socket_connection_and_close(client_socket, result_dict)
                    self._logger.info("Cloudlinux WPOS daemon reloaded for user %s", username)
                except (socket.error, AttributeError, struct.error):
                    self._logger.exception("Socket connection processing error", exc_info=True)
                except KeyError:
                    self._logger.exception("User not found during processing socket connection", exc_info=True)
            # Remove processed item from list
            self._socket_request_list.pop(0)
        # Reload redices for rest users
        for username in username_list_to_process:
            self._reload_redis_for_user(username)

    def _main_cycle(self):
        # Main daemon cycle
        i = 0
        while not self._is_terminate:
            try:
                # Process user requests from socket. Fills self._socket_queries_list
                self._process_socket_connections()
                # Self-reload if lvectl destroy was called
                self._force_reload()
                # Process all requests - start/stop redis for users
                self._process_all_requests()
                if self._is_terminate:
                    break
                time.sleep(1)
                i += 1
                if i > self._monitoring_interval:
                    # monitoring of existing redises
                    self._redises_monitoring()
                    i = 0
            except Exception:
                self._logger.exception("Cloudlinux WPOS daemon general error", exc_info=True)

    def run(self):
        """
        Main work daemon function
        """
        if get_process_pid(self._PID_FILENAME) is not None:
            # Error - daemon already works
            self._logger.warning("Cloudlinux WPOS daemon already works")
            return
        if os.path.exists(WPOS_DAEMON_SOCKET_FILE):
            # Error - daemon's socket already exist, remove it
            self._remove_socket()
        # Create daemon's socket
        try:
            self._socket = _create_socket()
        except (OSError, IOError) as e:
            message = "Can't create WPOS daemon socket. Error is: %s"
            # We used .warning instead .error because .error will send message to sentry
            self._logger.warning(message, str(e))
            return
        try:
            kill_all_users_redises(self._logger)
        except Exception as e:
            self._logger.exception(e)
        self._logger.info("Cloudlinux WPOS daemon started")
        self._setup_signals()
        write_pid_file(self._PID_FILENAME)
        # Force reload at start
        self.reload()
        # Main daemon_work
        self._main_cycle()
        self._remove_socket()
        # Terminate all user's redices
        kill_all_users_redises(self._logger)
        remove_pid_file(self._PID_FILENAME)
        self._logger.info("Cloudlinux WPOS daemon terminated")

    def _force_reload(self):
        """
        Force reload daemon
        """
        # Force reload if flag file exists
        if os.path.isfile(self._WPOS_DAEMON_RELOAD_FILE):
            self.reload()
            try:
                os.remove(self._WPOS_DAEMON_RELOAD_FILE)
            except (OSError, IOError):
                pass

    def _remove_socket(self):
        """
        Remove daemon's socket file
        """
        try:
            self._logger.info("Removing socket %s", WPOS_DAEMON_SOCKET_FILE)
            if self._socket is not None:
                self._socket.close()
            os.remove(WPOS_DAEMON_SOCKET_FILE)
            self._logger.info("Socket %s removed", WPOS_DAEMON_SOCKET_FILE)
        except (OSError, IOError) as e:
            self._logger.warning("Can't remove socket %s. Error: %s", WPOS_DAEMON_SOCKET_FILE, str(e))

    def stop(self):
        """
        Stops a working daemon process
        """
        pid = get_process_pid(self._PID_FILENAME)
        if pid:
            # Try killing the daemon process
            try:
                os.kill(pid, signal.SIGTERM)
            except OSError:
                # No such process
                pass
        # Terminate all user's redices
        kill_all_users_redises(self._logger)
        # Remove PID file
        remove_pid_file(self._PID_FILENAME)
        self._logger.info("Cloudlinux WPOS daemon stopped")

    def _get_user_redis_status(self, uid) -> dict:
        """
        Get redis status for user
        :param uid: User uid to reload
        :return: Dict to send to clwpos-user via socket
        """
        try:
            username = pwd.getpwuid(uid).pw_name
            is_redis_alive, result_dict = is_user_redis_alive(self._logger, username)
            # check error
            if result_dict["result"] != "success":
                return result_dict
            answer = {'result': 'success', 'timestamp': time.time(), 'status': is_redis_alive}
        except KeyError:
            answer = {'result': 'No such user', 'timestamp': time.time()}
        return answer

    def _validate_socket_connection(self, client_socket_obj: socket.socket) -> Tuple[bool, int, Optional[dict]]:
        """
        Validate socket connection. Check:
         - root connections;
         - connection user presense
        :return: tuple(error_flag, uid, user_request_dict)
          error_flag:
           True - Socket connection valid, should be processed. uid and user_request_dict filled
           False - invalid connection, should be skipped. uid == -1, user_request_dict = None
        """
        _uid = socket_utils.get_uid_from_socket(client_socket_obj)
        if _uid == 0:
            # root connections are disabled
            result: dict = {'result': 'Cloudlinux WPOS daemon socket connections are disabled for root'}
            socket_utils.send_dict_to_socket_connection_and_close(client_socket_obj, result)
            return False, -1, None
        if not is_uid_present(_uid):
            socket_utils.send_dict_to_socket_connection_and_close(client_socket_obj,
                                                                  {'result': 'No such user',
                                                                   'timestamp': time.time()})
            return False, -1, None
        user_request: dict = socket_utils.read_unpack_response_from_socket(client_socket_obj)
        if 'command' not in user_request:
            # User sent invalid query
            socket_utils.send_dict_to_socket_connection_and_close(client_socket_obj,
                                                                  {'result': 'Daemon received malformed query (command absent)',
                                                                   'timestamp': time.time()})
            return False, -1, None
        if user_request['command'] not in self._DAEMON_VALID_COMMANDS:
            # User sent invalid query
            socket_utils.send_dict_to_socket_connection_and_close(client_socket_obj,
                                                                  {'result': 'Daemon received invalid command',
                                                                   'timestamp': time.time()})
            return False, -1, None
        return True, _uid, user_request

    def _process_socket_connection(self, client_socket_obj: socket.socket):
        """
        Process sockect connection
        :param client_socket_obj: Client socket connection
        """
        is_connection_valid, uid, user_request = self._validate_socket_connection(client_socket_obj)
        if not is_connection_valid:
            return
        if user_request['command'] == self._DAEMON_GET_REDIS_STATUS_COMMAND:
            # Get redis status for user
            result: dict = self._get_user_redis_status(uid)
            socket_utils.send_dict_to_socket_connection_and_close(client_socket_obj, result)
            return
        # Add socket and other data for further process
        self._socket_request_list.append(
            Socket_request(client_socket=client_socket_obj, client_uid=uid,
                           client_cmd_dict=user_request))

    def _process_socket_connections(self):
        """
        Process socket connections
        """
        socket_list = [self._socket]
        readable, _, _ = select.select(socket_list, [], [], 1)
        for _sock_object in readable:
            try:
                client_socket_obj, _ = _sock_object.accept()
                self._process_socket_connection(client_socket_obj)
            except (socket.error, json.JSONDecodeError, AttributeError, struct.error, UnicodeDecodeError):
                self._logger.exception("Socket connection processing error", exc_info=True)


def daemon_main(_opts):
    daemon = WposDaemon()
    if _opts.action == "start":
        daemon.run()
    if _opts.action == "restart":
        daemon.stop()
        daemon.run()
    elif _opts.action == "stop":
        daemon.stop()
    elif _opts.action == "reload":
        daemon.reload()
