# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#

# pylint: disable=no-absolute-import

import logging
import os
import signal
import json

from clcommon.utils import get_process_pid
from clwpos.logsetup import setup_logging, init_wpos_sentry_safely


class WposDaemonBase:
    """
    WPOS daemon base class. Signals, signal handlers, setup daemon logger, pid file operations
    """
    _PID_FILENAME = "/var/run/clwpos_monitoring.pid"
    _DAEMON_LOGFILE_PATH = "/var/log/clwpos/daemon.log"
    _DAEMON_CONFIG_PATH = "/etc/clwpos/daemon_conf.json"
    # Default redis monitoring interval in 5 sec units.
    # 60 * 5 sec = 300 seconds = 5 min
    _DEFAULT_MONITORING_INTERVAL = 60

    def __init__(self):
        self._logger = setup_logging(
            caller_name=__name__,
            console_level=logging.INFO,
            logfile_path=self._DAEMON_LOGFILE_PATH
        )
        init_wpos_sentry_safely(self._logger)
        self._reload_config_need = False
        self._is_terminate = False
        self._monitoring_interval = self._DEFAULT_MONITORING_INTERVAL

    def _update_monitoring_interval(self):
        """
        Update redis monitoring interval in class variable
        """
        try:
            with open(self._DAEMON_CONFIG_PATH, 'r') as f:
                daemon_config = json.load(f)
                self._monitoring_interval = daemon_config['monitoring_interval_5secs'] * 5
        except (OSError, IOError, json.JSONDecodeError, KeyError):
            # Config read error, stay default
            pass

    def _setup_signals(self):
        """
        Setup daemon signal handlers
        """
        # Setup SIGHUP signal handler for config reload
        signal.signal(signal.SIGHUP, self._sighup_handler)
        # Setup Ctrl-C signal handler. Used for console debug
        signal.signal(signal.SIGINT, self._sigint_handler)
        # Setup SIGTERM signal. Called when daemon stops
        signal.signal(signal.SIGTERM, self._sigterm_handler)

    def run(self):
        """
        Main work daemon function (implement in child class)
        """
        raise NotImplementedError()

    def stop(self):
        """
        Stops a working daemon process (implement in child class)
        """
        raise NotImplementedError()

    def reload(self):
        """
        Reload working daemon process by sending SIGHUP signal to it
        """
        try:
            pid = get_process_pid(self._PID_FILENAME)
            if pid:
                os.kill(pid, signal.SIGHUP)
        except:
            pass

    def _sighup_handler(self, signum, frame):
        """
        SIGHUP signal handler
        """
        # Load new user configuration from config file
        self._reload_config_need = True
        self._logger.info("SIGHUP: Cloudlinux WPOS daemon sheduled for reload")

    def _sigint_handler(self, signum, frame):
        """
        SIGINT (Ctrl-C) signal handler. For console debug
        """
        self._logger.info("Ctrl-C received, exit")
        # Terminate main cycle
        self._is_terminate = True

    def _sigterm_handler(self, signum, frame):
        """
        SIGTERM signal handler. Called when daemon stopping
        """
        # Terminate main cycle
        self._is_terminate = True
