# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#

# Redis manipulation library for Cloudlinux WPOS daemon

# pylint: disable=no-absolute-import

import pwd
import os
import traceback
import subprocess
import signal
import shutil
import time
from logging import Logger
from typing import List, Optional, Tuple

from secureio import write_file_via_tempfile
from clcommon.clpwd import drop_privileges
from clcommon.utils import run_command, ExternalProgramFailed, is_user_present
from clcommon.cpapi import cpusers, userdomains
from clcommon.lib.cledition import is_cl_solo_edition
from clcommon.clpwd import ClPwd

from clwpos import gettext as _
from clwpos.cl_wpos_exceptions import WposError, WpUserMissing
from clwpos.utils import USER_WPOS_DIR, demote
from clwpos.user.config import UserConfig, ConfigError
from clwpos.optimization_modules import get_allowed_modules
from clcommon.cpapi.cpapiexceptions import NoPackage


_USER_REDIS_CONF_PATTERN = """# !!! WARNING !!! AUTO-GENERATED FILE, PLEASE DO NOT MODIFY IT
maxclients 16
databases 16
maxmemory-policy allkeys-lru
appendonly no
appendfsync always
port 0
unixsocket {socket_path}
unixsocketperm 600
dir {clwpos_dir}
maxmemory {maxmemory}
save ""
"""
_REDIS_SERVER_BIN_FILE = '/opt/alt/redis/bin/redis-server'
_REDIS_CLI_BIN_FILE = '/opt/alt/redis/bin/redis-cli'


def _write_redis_config_for_user(username: str, user_homedir: str, clwpos_dir: str, redis_socket_path: str,
                                 redis_config_path: str, maxmemory: str):
    """
    Writes redis config for user
    :param username: User name for write config
    :param user_homedir: User's homedir
    :param redis_socket_path: Full path to user's redis socket
    :param redis_config_path: Full path to user's redis config file
    :param maxmemory: maxmemory value to write to redis config
    :return: None
    """
    redis_config_content = _USER_REDIS_CONF_PATTERN.format(socket_path=redis_socket_path,
                                                           clwpos_dir=clwpos_dir, maxmemory=maxmemory)
    try:
        with drop_privileges(username):
            # We are user here
            try:
                clwpos_dir = os.path.join(user_homedir, USER_WPOS_DIR)
                try:
                    os.makedirs(clwpos_dir, exist_ok=True)
                except (OSError, IOError,) as system_error:
                    raise WposError(message=_("Can't create directory %(directory)s. "),
                                    details=_('The operating system reported error "%(system_error)s".'),
                                    context={"directory": clwpos_dir, 'system_error': system_error})
                # Write redis config
                write_file_via_tempfile(redis_config_content, redis_config_path, 0o600)
            except (OSError, IOError) as system_error:
                raise WposError(message=_("Error happened while writing redis config for user '%(user)s' "),
                                details=_('The operating system reported error "%(system_error)s".'),
                                context={"user": username, "system_error": system_error})
    except ClPwd.NoSuchUserException:
        raise WpUserMissing(username)


def _remove_clwpos_dir_for_user(_logger: Logger, username: str):
    """
    Remove /home/<user>/.clwpod dir with all content
    :param username: User name to remove dir
    """
    try:
        with drop_privileges(username):
            # We are user here
            try:
                user_pwd = pwd.getpwnam(username)
                # /home/username/.clwpos/
                shutil.rmtree(os.path.join(user_pwd.pw_dir, USER_WPOS_DIR))
            except (OSError, IOError):
                _logger.exception("Can't remove .clwpos directory for user %s", username)
    except ClPwd.NoSuchUserException:
        raise WpUserMissing(username)


def _get_pids_for_file(file_path: str) -> List[int]:
    """
    Retrieves list of PID list processes, which uses file (using fuser utility)
    This can find any process (for example php), not only redis service process
    :param file_path: Filename to check
    :return: PID list
    """
    try:
        # # /usr/sbin/fuser /home/cltest1/.clwpos/redis.sock
        # /home/cltest1/.clwpos/redis.sock: 55882 [105766 251507]
        std_out = run_command(['/sbin/fuser', file_path], return_full_output=False)
        lines_list = std_out.split('\n')
        # Get PID list from output
        s_pid_list = lines_list[0].split(':')[1].strip()
        pid_list = []
        for s_pid in s_pid_list.split(' '):
            try:
                pid_list.append(int(s_pid.strip()))
            except ValueError:
                pass
        return pid_list
    except (ExternalProgramFailed, IndexError):
        pass
    return []


def _get_user_pids(username: str) -> List[int]:
    """
    Update PID list in cache for user using /bin/ps utility
    :param: username: Username to scan
    :return: None
    """
    # /bin/ps -o"pid" -u cltest1
    # PID
    # 1608661
    # 1638657
    # ......
    # Get user's PID list
    try:
        std_out = run_command(['/bin/ps', '-o', 'pid', '-u', username], return_full_output=False)
    except ExternalProgramFailed:
        return []
    lines_list = std_out.split('\n')
    if len(lines_list) < 2:
        return []
    # Remove header line
    user_pid_list = []
    lines_list = lines_list[1:]
    for line in lines_list:
        line = line.strip()
        if line:
            try:
                user_pid_list.append(int(line.strip()))
            except ValueError:
                pass
    return user_pid_list


def _get_user_redis_pids(username: str, home_dir: str) -> Optional[List[int]]:
    """
    Get redis PID list for user
    :param username: user name
    :param home_dir: User's  homedir
    :return: PID list or [] if user has no redis
    """
    redis_socket_file = os.path.join(home_dir, USER_WPOS_DIR, 'redis.sock')
    pid_list_sock = _get_pids_for_file(redis_socket_file)
    user_pids = _get_user_pids(username)
    pid_list = []
    for pid in pid_list_sock:
        if pid in user_pids:
            pid_list.append(pid)
    return pid_list


def _get_valid_docroots_from_config(logger: Logger, username: str, user_homedir: str,
                                    docroot_list_to_validate: List[str]) -> List[str]:
    """
    Validates docroots from list
    :param logger: Logger to log errors
    :param username: User name
    :param user_homedir: User's homedir
    :param docroot_list_to_validate: Docroot list to validate
    :return: List of valid docroots
    """
    if not docroot_list_to_validate:
        # Nothing to validate
        return []
    try:
        # Get active docroots for user from panel
        userdomains_data_list = userdomains(username)
    except (OSError, IOError, IndexError, NoPackage) as e:
        # Skip all cpapi.userdomains errors
        logger.warning("Can't get user list from panel: %s", str(e))
        return []
    # Add some filename to all docroots to avoid any problems with / at the end of paths
    docroots_from_panel = []
    for _, document_root in userdomains_data_list:
        docroots_from_panel.append(os.path.join(document_root, '1'))
    # Validate docroots from config
    valid_docroot_list: List[str] = []
    for _dr in docroot_list_to_validate:
        tempname = os.path.join(user_homedir, _dr, '1')
        if tempname in docroots_from_panel:
            valid_docroot_list.append(_dr)
    return valid_docroot_list


def _is_redis_plugin_enabled(logger: Logger, username: str, user_homedir: str, user_config_dict: dict) -> bool:
    """
    Get redis status according to user's and admin's configs
    :param logger: Logger to log errors
    :param username: User name
    :param user_homedir: User's homedir
    :param user_config_dict: User's config
    :return: True - module enabled, False - disabled
    """
    # TODO: Refactor this function in LU-2613
    # # Admin's config example:
    # allowed_modules_list example: ['object_cache']
    allowed_modules_list = get_allowed_modules()
    if not allowed_modules_list:
        return False
    # WPOS enabled by admin, check user's config
    # User's config example:
    # {
    #     "docroots": {
    #         "public_html": {"": ["object_cache"], "1": []},
    #         "public_html/gggh.com": {"wp1": ["object_cache"], "wp2": []}
    #     },
    #     "maxmemory": "512mb"
    # }
    docroots_to_validate: list = list(user_config_dict['docroots'].keys())
    valid_wp_docroots = _get_valid_docroots_from_config(logger, username, user_homedir, docroots_to_validate)
    for wp_docroot in valid_wp_docroots:
        wp_data = user_config_dict['docroots'][wp_docroot]
        for module_list in wp_data.values():  # By WP paths
            if isinstance(module_list, list) and len(module_list):
                # Redis enabled
                return True
    return False


def _kill_process_by_pid(_logger: Logger, pid: int):
    """
    Kill process by pid
    :param _logger: Logger to log errors
    :param pid: Process pid to kill
    """
    try:
        os.kill(pid, signal.SIGTERM)  # 15
        os.waitpid(pid, os.WNOHANG)
        time.sleep(5)
        os.kill(pid, signal.SIGKILL)  # 9
        os.waitpid(pid, os.WNOHANG)
    except OSError as e:
        # No such process
        _logger.warning("Can't kill redis process: %s", str(e))


def _kill_all_redises_for_user(logger: Logger, username: str):
    """
    Kill all user's redice processes
    :param logger: Logger to log errors
    :param username: User name
    """
    if not is_user_present(username):
        return
    user_pwd = pwd.getpwnam(username)
    redis_pid_list = _get_user_redis_pids(user_pwd.pw_name, user_pwd.pw_dir)
    for redis_pid in redis_pid_list:
        _kill_process_by_pid(logger, redis_pid)


def remove_clwpos_dir_for_all_users(logger: Logger):
    """
    Remove .clwpos dir for all panel users
    :param logger: Logger to log errors
    """
    try:
        users = cpusers()
    except (OSError, IOError, IndexError, NoPackage) as e:
        # Log all cpapi.userdomains errors
        logger.warning("Can't get user list from panel: %s", str(e))
        return
    for username in users:
        try:
            _remove_clwpos_dir_for_user(logger, username)
        except WposError as e:
            # Skip user not found error
            logger.warning("User %s present in panel but absent in system. Error is %s", username, str(e))


def kill_all_users_redises(logger: Logger):
    """
    Find and kill lost redices for all panel users
    :param logger: Daemon's logger
    """
    try:
        users = cpusers()
    except (OSError, IOError, IndexError, NoPackage) as e:
        logger.warning("Can't get user list from panel: %s", str(e))
        return
    for username in users:
        _kill_all_redises_for_user(logger, username)


def is_user_redis_alive(_logger: Logger, username: str) -> Tuple[bool, dict]:
    """
    Check user's redis is alive
    :param _logger: Daemon's logger
    :param username: Username to check redis process
    return True/False - redis alive/not alive
    :return: Tuple:
      If redis is working/not working - (True/False, {"result": "success"})
      error  - (False, {"result": "error", "context": "..."})
    """
    # # /opt/alt/redis/bin/redis-cli -s /home/cltest1/.clwpos/redis.sock ping
    # Could not connect to Redis at /home/cltest1/.clwpos/redis.sock: No such file or directory
    # # echo $?
    # 1
    # # /opt/alt/redis/bin/redis-cli -s /home/cltest1/.clwpos/redis.sock ping
    # PONG
    # # echo $?
    # 0
    try:
        user_pwd = pwd.getpwnam(username)
        redis_socket_path = os.path.join(user_pwd.pw_dir, USER_WPOS_DIR, 'redis.sock')
        if is_cl_solo_edition(skip_jwt_check=True):
            # CL Solo
            cmd = [_REDIS_CLI_BIN_FILE, '-s', redis_socket_path, 'ping']
        else:
            cmd = ['/sbin/cagefs_enter_user', '--no-fork', username,
                   _REDIS_CLI_BIN_FILE, '-s', redis_socket_path, 'ping']
        proc = subprocess.Popen(cmd, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL,
                                preexec_fn=demote(user_pwd.pw_uid, user_pwd.pw_gid),
                                cwd=user_pwd.pw_dir)
        proc.communicate()
        return proc.returncode == 0, {"result": "success"}
    except (ExternalProgramFailed, KeyError, ClPwd.NoSuchUserException, IOError, OSError) as e:
        str_exc = traceback.format_exc()
        _logger.warning("Redis check error for user %s. Error is: %s", username, str_exc)
        return False, {"result": "Redis server start error %(error)s",
                       "context": {"error": str(e)}}


def _start_redis_server_for_user(_logger: Logger, username: str, maxmemory: str) -> Tuple[int, dict]:
    """
    Start redis server for supplied user
    :param _logger: Daemon's logger
    :param username: Username to setup redis
    :param maxmemory: maxmemory value to write to redis config
    :return: Tuple:
      If redis was started for user - (PID of new redis process, {"result": "success"})
      else - redis was not started  - (-1, {"result": "error", "context": "...."})
    """
    user_pwd = pwd.getpwnam(username)
    # /home/username/.clwpos/
    wpos_dir = os.path.join(user_pwd.pw_dir, USER_WPOS_DIR)
    redis_socket_path = os.path.join(wpos_dir, 'redis.sock')
    redis_config_path = os.path.join(wpos_dir, 'redis.conf')
    _write_redis_config_for_user(username, user_pwd.pw_dir, wpos_dir, redis_socket_path, redis_config_path, maxmemory)
    # start redis
    # /sbin/cagefs_enter_user --no-fork <username> /usr/bin/redis-server <username_home>/.clwpos/redis.conf
    try:
        if is_cl_solo_edition(skip_jwt_check=True):
            # CL Solo
            proc = subprocess.Popen([_REDIS_SERVER_BIN_FILE, redis_config_path],
                                    preexec_fn=demote(user_pwd.pw_uid, user_pwd.pw_gid),
                                    shell=False, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        else:
            # CL Shared
            proc = subprocess.Popen(['/sbin/cagefs_enter_user', '--no-fork', username,
                                     _REDIS_SERVER_BIN_FILE, redis_config_path],
                                    shell=False, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        return proc.pid, {"result": "success"}
    except (OSError, IOError) as e:
        str_exc = traceback.format_exc()
        _logger.warning("Redis server start error for user %s. Error is: %s", username, str_exc)
        return -1, {"result": "Redis server start error %(error)s",
                    "context": {"error": str(e)}}


def reload_redis_for_user(_logger: Logger, username: str, old_redis_pid: Optional[int]) -> Tuple[int, dict]:
    """
    Reloads redis for supplied user
    :param _logger: Daemon's logger
    :param username: Username to setup redis
    :param old_redis_pid: Old Redis PID for kill
    :return: Tuple:
      If redis was started for user - (PID of new redis process, {"result": "success"})
      else - redis was not started  - (-1, {"result": "error", "context": ""})
    """
    if old_redis_pid:
        _kill_process_by_pid(_logger, old_redis_pid)
    try:
        with drop_privileges(username):
            try:
                user_config_class = UserConfig(username)
                user_config = user_config_class.read_config()
            except ConfigError:
                # User's config read error
                _logger.warning("Can't reload redis for user %s. User's config read error", username)
                return -1, {"result": "Can't reload redis for user %(user)s. User's config read error",
                            "context": {"user": username}}
            user_homedir = pwd.getpwnam(username).pw_dir
            is_redis_enable = _is_redis_plugin_enabled(_logger, username, user_homedir, user_config)
            if is_redis_enable:
                # Start redis and return new PID
                return _start_redis_server_for_user(_logger, username,
                                                    user_config.get('max_cache_memory',
                                                                    UserConfig.DEFAULT_MAX_CACHE_MEMORY))
    except (ClPwd.NoSuchUserException, KeyError):
        _logger.warning("Can't reload redis for user '%s'. User not found.", username)
        return -1, {"result": "Can't reload redis for user '%(user)s'. User not found.",
                    "context": {"user": username}}
    # Redis is disabled by user's config
    return -1, {"result": "success"}
