# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2020 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

from __future__ import absolute_import

import time
import json
import os

from typing import Dict, Union, List, Tuple
from secureio import write_file_via_tempfile
from clcommon.clpwd import ClPwd
from clwpos.utils import acquire_lock
from clwpos.optimization_modules import get_all_modules_data
from clwpos.logsetup import setup_logging
from clwpos.cl_wpos_exceptions import WposError

from clwpos.constants import (
    SCANNING_STATUS,
    LAST_SCANNED_TS,
    SCAN_CACHE,
    CLWPOS_VAR_DIR,
    ALLOWED_MODULES_JSON
)


class ReportGeneratorError(WposError):
    """
    Raised when some bad happened during report generating/getting
    """
    pass


class ReportGenerator:
    """
    Class for admin report generation/getting
    """

    def __init__(self):
        self._clpwd = ClPwd()
        self.logger = setup_logging(__name__)

    @staticmethod
    def scan():
        # just a stub, @ikartavcev will replace this
        my_data = {
            'stackoverflow': {
                'wp_sites_count': '20'
            },
            'puser': {
                'wp_sites_count': '4'
            },
            'ncuser': {
                'wp_sites_count': '2'
            }
        }
        write_file_via_tempfile(json.dumps({'users': my_data}), SCAN_CACHE, 0o600)
        write_file_via_tempfile(str(int(time.time())), LAST_SCANNED_TS, 0o600)
        write_file_via_tempfile('1/222', SCANNING_STATUS, 0o600)
        return {
            'total_users_scanned': 0,
            'total_users_to_scan': 26
        }

    def get(self, target_users=None):
        """
         1. get-report when there is no cache file:
          - start generating report
          - return total_users_scanned, total_users_to_scan keys
        2. get-report when there is no cache file, but scanning
           in progress:
           - return total_users_scanned, total_users_to_scan keys
        3. get-report when cache present, no scanning running
          - return data from cache
          - no keys total_users_scanned, total_users_to_scan in response
        4. get-report when cache present, scanning is running
          - return data from cache
          - return total_users_scanned, total_users_to_scan keys
        """
        report = {}
        scanned_cache = self._get_from_file_with_locking(SCAN_CACHE, is_json=True)
        scanning_status = self._get_from_file_with_locking(SCANNING_STATUS, is_json=False, with_lock=True)

        # no cache and scanning is not running right now ->
        # run scanning
        if not scanned_cache and not scanning_status:
            scan = self.scan()
            return scan

        # no cache, but scanning is running right now ->
        # returns status from status file
        if scanning_status:
            already_scanned, total_to_scan = self._parse_scanning_status(scanning_status)
            report['total_users_scanned'] = already_scanned
            report['total_users_to_scan'] = total_to_scan

        # if scanned cache -> append it to report
        if scanned_cache:
            last_scanned_ts = self._get_from_file_with_locking(LAST_SCANNED_TS, is_json=False)
            report['last_scan_time'] = last_scanned_ts

            # cache already generated, let`s read it
            users_cache = scanned_cache['users']
            report['users'], report['total_wordpress_count'], report['total_users_count'], \
                report['total_users_count_enabled_cache'] = self._get_users_report(users_cache, target_users)
        return report

    def _get_users_report(self, user_scanned_data: Dict, target_users: Union[None, List]) -> Tuple[List, int, int, int]:
        """
        Returns list of final users data:
        example:
        [
        {"username": "user1", "modules": {"object_cache": true}, "wp_sites_count": "1"},
        ....
        ]

        """
        all_users_info = []
        total_wp_sites_count, total_users_count, total_users_with_enabled_cache = 0, 0, 0
        for username, scanned_data in user_scanned_data.items():
            if target_users and username not in target_users:
                continue
            uid = self._uid_by_name(username)
            if not uid:
                self.logger.warning('Cannot get user %s uid', username)
                continue
            allowed_modules_path = os.path.join(CLWPOS_VAR_DIR, 'uids', str(uid), ALLOWED_MODULES_JSON)

            # let`s not break report collection for all users
            try:
                modules_data = get_all_modules_data(allowed_modules_path)
            except Exception:
                self.logger.exception('Unable to get modules data from %s', allowed_modules_path)
                modules_data = {}

            all_users_info.append({
                'username': username,
                'modules': modules_data,
                'wp_sites_count': scanned_data['wp_sites_count']
            })
            total_users_count += 1
            total_wp_sites_count += int(scanned_data['wp_sites_count'])
            if modules_data.get('object_cache'):
                total_users_with_enabled_cache += 1

            # stop iterating through users if targets already found
            if target_users and len(target_users) == len(all_users_info):
                break

        return all_users_info, total_wp_sites_count, total_users_count, total_users_with_enabled_cache

    def _get_from_file_with_locking(self, path: str, is_json: bool, with_lock: bool = False) -> Union[Dict, str, None]:
        """
        Reads file with locking if needed
        """
        if not os.path.exists(path):
            return None
        try:
            if with_lock:
                with acquire_lock(path):
                    return self._read_report_file(path, is_json)
            else:
                return self._read_report_file(path, is_json)
        except Exception as e:
            raise ReportGeneratorError(message='Can`t read (file)%s. Reason: (error)%s',
                                       context={'file': path, 'error': str(e)})

    @staticmethod
    def _read_report_file(path: str, is_json: bool) -> Union[Dict, str, None]:
        """
        Reads file for report
        """
        with open(path, "r") as file:
            if is_json:
                return json.load(file)
            else:
                return file.read().strip()

    def _uid_by_name(self, name):
        try:
            return self._clpwd.get_uid(name)
        except ClPwd.NoSuchUserException:
            return None

    @staticmethod
    def _parse_scanning_status(status):
        try:
            scanning_data = status.split('/')
            already_scanned, total_to_scan = int(scanning_data[0]), int(scanning_data[1])
        except Exception as e:
            raise ReportGeneratorError(message='Cannot parse content of status file %(content)s. Error: %(error)s.',
                                       context={'content': status, 'error': str(e)})
        return already_scanned, total_to_scan
