#!/opt/alt/python37/bin/python3 -bb
# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#

from __future__ import absolute_import

import json
import socket
import struct
import pwd
from typing import Optional


# uint32_t, big endian
_format = '>I'


def get_uid_from_socket(sock_object: socket.socket) -> int:
    """
    Retrieve credentials from SO_PEERCRED option
    :param sock_object: Socket object
    :return: uid of user, which connects to this socket.
    """
    _format = '3i'
    creds = sock_object.getsockopt(socket.SOL_SOCKET, socket.SO_PEERCRED, struct.calcsize(_format))
    # creds contains _pid, _uid, _gid - 3 uint32_t numbers
    _, _uid, _ = struct.unpack(_format, creds)
    # Check user existance
    _ = pwd.getpwuid(_uid).pw_name
    return _uid


def pack_data_for_socket(data_dict: dict) -> bytes:
    """
    Prefix message with a 4-byte length
    :param data_dict: Data dict for send
    :return: byte array for send to socket
    """
    msg_bytes = json.dumps(data_dict).encode('utf-8')
    # Output data format:
    # 4 bytes unsigned int, big-endian - data_len
    # data_len                         - data_bytes
    return struct.pack(_format, len(msg_bytes)) + msg_bytes


def read_unpack_response_from_socket(sock_object: socket.socket) -> dict:
    """
    Read length-prefixed amount of data from socket
    :param sock_object: Socket object to read data
    :return: received from socket dictionary
    """
    # Socket Input data format:
    # 4 bytes unsigned int, big-endian - data_len
    # data_len                         - data_bytes
    chunk = 4096
    raw_msglen = sock_object.recv(4)
    msglen = struct.unpack(_format, raw_msglen)[0]
    msg = bytes()
    while len(msg) != msglen:
        msg += sock_object.recv(chunk)
    return json.loads(msg.decode('utf-8'))


def send_dict_to_socket_connection_and_close(connection: socket.socket, data_to_send: dict):
    """
    Sends dictionary to socket connection and close it
    :param connection: Socket connection to send data
    :param data_to_send: Data dict to send
    """
    bytes_to_send = pack_data_for_socket(data_to_send)
    connection.sendall(bytes_to_send)
    connection.close()
