# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2020 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

# wposctl.py - work code for clwposctl utility

from __future__ import absolute_import

import argparse
import json
import os
import shutil
import subprocess
import sys
from copy import deepcopy
from typing import Dict, Iterator, Set, Tuple

import cpanel
from clcommon import cpapi
from clwpos.optimization_modules import (
    ALL_OPTIMIZATION_MODULES,
    ALLOWED_MODULES_CONFIG_VERSION,
    ALLOWED_MODULES_CONFIG,
    get_all_modules_data
)
from clcommon.clpwd import drop_privileges
from clwpos.cl_wpos_exceptions import WposError

from clwpos.user.config import UserConfig

from clwpos import gettext as _
from cpanel import enable_without_config_affecting, disable_without_config_affecting, DocRootPath
from clwpos.parse import ArgumentParser, CustomFormatter
from clwpos.logsetup import setup_logging, init_wpos_sentry_safely
from clwpos.utils import catch_error, error_and_exit, print_data, check_license_decorator
from clcommon.lib.cledition import is_cl_solo_edition
from clcommon.cpapi.cpapiexceptions import NoPackage
from clwpos.report_generator import ReportGenerator, ReportGeneratorError
from clwpos.utils import acquire_lock

DISABLED_OMS_MESSAGE = _("All optimization modules are currently disabled. "
                         "End-user WPOS interface blocked.")

parser = ArgumentParser(
    "/usr/bin/clwpos-admin",
    "Utility for control CL WPOS admin interface",
    formatter_class=CustomFormatter
)

_logger = setup_logging(__name__)


class CloudlinuxWposAdmin(object):
    """
    Class for run cloudlinux-wpos-admin commands
    """

    def __init__(self):
        self._is_json = False
        self._opts: argparse.Namespace
        self._logger = setup_logging(__name__)
        init_wpos_sentry_safely(self._logger)
        self.clwpos_path = "/var/clwpos"
        self.modules_allowed_name = "modules_allowed.json"
        self.is_solo = is_cl_solo_edition(skip_jwt_check=True)

    @catch_error
    def run(self, argv):
        """
        Run command action
        :param argv: sys.argv[1:]
        :return: clwpos-user utility retcode
        """
        self._parse_args(argv)
        result = getattr(self, self._opts.command.replace("-", "_"))()
        print_data(self._is_json, result)

    def _parse_args(self, argv):
        """
        Parse command line arguments
        :param argv: sys.argv[1:]
        """
        self._opts = parser.parse_args(argv)
        self._is_json = True

    @parser.command(help="Uninstall cache for all domain during downgrade")
    def uninstall_cache_for_all_domains(self) -> dict:
        """
        This command used during downgrade to lve-utils, which version does not support clwpos
        :return:
        """
        try:
            users = cpapi.cpusers()
        except (OSError, IOError, IndexError, NoPackage) as e:
            self._logger.warning("Can't get user list from panel: %s", str(e))
            return {}
        for username in users:
            with drop_privileges(username):
                for doc_root, wp_path, module in _enabled_modules(username):
                    disable_without_config_affecting(DocRootPath(doc_root), wp_path, module=module)
        return {}

    @parser.argument(
        "--modules",
        help="Argument for module of list of comma separated modules",
        type=str,
        choices=ALL_OPTIMIZATION_MODULES,
        required=True
    )
    @parser.mutual_exclusive_group(
        [
            (["--all"], {"help": "Argument for all users in the panel", "action": "store_true"}),
            (["--users"], {"help": "Argument for user or list of comma separated users", "type": str}),
        ],
        required=True,
    )
    @parser.argument("--allowed", help="Enable module if present", action="store_true")
    @parser.argument("--disallowed", help="Disable module if present", action="store_true")
    @parser.command(help="Managing list of allowed modules for users")
    @check_license_decorator
    def set_module(self) -> dict:
        """
        Write info related to module allowance into user file
        """
        if self.is_solo:
            user = "solo"
        else:
            error_and_exit(
                self._is_json, {"result": _("Only CloudLinux OS Solo edition is supported.")}
            )
            # code for shared implementation
            # user = [user.strip() for user in self.opts.users]
        modules = [module.strip() for module in self._opts.modules.split(",")]
        if self._opts.allowed:
            allowed_state = True
        elif self._opts.disallowed:
            allowed_state = False
        else:
            error_and_exit(
                self._is_json, {"result": 'Missing "--allowed" or "--disallowed" option.'}
            )

        if self.is_solo:
            modules_allowed_dir = os.path.join(self.clwpos_path, user)
            modules_allowed_path = os.path.join(
                modules_allowed_dir, self.modules_allowed_name
            )
        else:
            # path in the shared would differ from solo
            # /var/clpwos/uids/{uid}/modules_allowed.json
            error_and_exit(
                self._is_json, {"result": "Non-solo edition is not supported."}
            )
        with acquire_lock(modules_allowed_path):
            if os.path.isfile(modules_allowed_path):
                with open(modules_allowed_path, "r") as f:
                    data = json.loads(f.read())

                try:
                    old_state = deepcopy(data["modules"])
                    data["modules"].update(dict.fromkeys(modules, allowed_state))
                except KeyError:
                    data = {
                        "version": str(ALLOWED_MODULES_CONFIG_VERSION),
                        "modules": dict.fromkeys(modules, allowed_state),
                    }
                new_state = deepcopy(data["modules"])
            else:
                if not os.path.isdir(modules_allowed_dir):
                    os.mkdir(modules_allowed_dir, 0o755)
                old_state = dict.fromkeys(ALL_OPTIMIZATION_MODULES, True)
                data = {
                    "version": str(ALLOWED_MODULES_CONFIG_VERSION),
                    "modules": dict.fromkeys(modules, allowed_state),
                }
                new_state = deepcopy(data["modules"])

            json_data = json.dumps(data, indent=4)

            try:
                with open(modules_allowed_path, "w") as f:
                    f.write(json_data)
            except (IOError, OSError) as err:
                raise WposError(
                    message=_("Configuration file '%(path)s' update failed."),
                    details=str(err),
                    context=dict(path=modules_allowed_path)
                )
            synchronize_plugins_status(old_state, new_state)

            if self.all_modules_disabled(data["modules"]):
                warning_dict = {"warning": DISABLED_OMS_MESSAGE}
            else:
                warning_dict = {}

            if allowed_state:
                subprocess.Popen(['/sbin/configure_redis_extension'])

            return warning_dict

    @catch_error
    @parser.argument("--hide-icon", help="", type=str)
    @parser.argument("--show-icon", help="", type=str)
    @parser.command(help="Managing options")
    @check_license_decorator
    def set_options(self):
        pass

    @catch_error
    @parser.mutual_exclusive_group(
        [
            (["--all"], {"help": "Argument for all users in the panel", "action": "store_true"}),
            (["--users"], {"help": "Argument for user or list of comma separated users", "type": str}),
        ],
        required=True,
    )
    @parser.command(help="Return the report about allowed and restricted user's modules")
    def get_report(self) -> dict:
        """
        Print report in stdout.
        [!ATTENTION!] response jsons are different for Solo and Shared!
        """
        # TODO: enable this check for Shared after beta release
        if not self.is_solo:
            error_and_exit(
                self._is_json, {"result": "Non-solo edition is not supported."}
            )

        if self.is_solo:
            try:
                modules = get_all_modules_data()
            except (KeyError, json.JSONDecodeError):
                raise WposError(
                    message=_("Configuration file '%(config_path)s' is corrupted. "
                              "Check it and make sure it has valid json format.\n"
                              "Contact CloudLinux support in case you need any assistance."),
                    context=dict(config_path=ALLOWED_MODULES_CONFIG)
                )
            report = {'modules': modules}
        else:
            try:
                users = self._opts.users.split(',') if self._opts.users else None
                report = ReportGenerator().get(target_users=users)
            except ReportGeneratorError as e:
                error_and_exit(
                    self._is_json,
                    {
                        'result': e.message,
                        'context': e.context
                    }
                )
            except Exception as e:
                error_and_exit(
                    self._is_json,
                    {
                        'result': f'Error during getting report: {e}'
                    }
                )
        return report

    @staticmethod
    def all_modules_disabled(modules: Dict[str, bool]) -> bool:
        """
        Check if all optimization modules are disabled.
        """
        if any(status is True for status in modules.values()):
            return False
        return True


def synchronize_plugins_status(old_state: dict, new_state: dict):
    """
    Compare old and new states of modules in admin's wpos config,
    determine what modules should be enabled and disabled
    and synchronize new state for each panel's user.
    """
    old_state = {key for key, value in old_state.items() if value}
    new_state = {key for key, value in new_state.items() if value}

    enabled_modules = new_state - old_state
    disabled_modules = old_state - new_state

    for username in cpapi.cpusers():
        synchronize_plugins_for_user(username, enabled_modules, disabled_modules)


def synchronize_plugins_for_user(
        username: str, enabled_modules: Set[str], disabled_modules: Set[str]):
    """
    Iterate through user's docroots and wp_paths
    and enable/disable modules with wp-cli
    not modifiing user's wpos config.
    """
    with drop_privileges(username):
        for doc_root, wp_path, _ in _enabled_modules(username):

            for module in enabled_modules:
                enable_without_config_affecting(
                    DocRootPath(doc_root),
                    wp_path,
                    module=module,
                )

            for module in disabled_modules:
                disable_without_config_affecting(
                    DocRootPath(doc_root),
                    wp_path,
                    module=module,
                )

            try:
                cpanel.reload_redis()
            except Exception as e:
                _logger.exception(e)


def _enabled_modules(username: str) -> Iterator[Tuple[str, str, str]]:
    return UserConfig(username=username).enabled_modules()
